import tensorflow.compat.v1 as tf
from collections import OrderedDict
import e2edutch.coref_model
import os
os.environ["CUDA_VISIBLE_DEVICES"] = "-1"

tf.disable_v2_behavior()

test_config = {'max_top_antecedents': 30,
               'max_training_sentences': 30,
               'top_span_ratio': 0.4,
               'filter_widths': [3, 4, 5],
               'filter_size': 50,
               'char_embedding_size': 8,
               'char_vocab_path': 'data/char_vocab.dutch.txt',
               'context_embeddings': OrderedDict([('path', 'test/data/test-embeddings.vec'),
                                                  ('size', 10),
                                                  ('cased', True)]),
               'head_embeddings': OrderedDict([('path', 'test/data/test-embeddings.vec'),
                                               ('size', 10),
                                               ('cased', True)]),
               'contextualization_size': 200,
               'contextualization_layers': 3,
               'ffnn_size': 150,
               'ffnn_depth': 2,
               'feature_size': 20,
               'max_span_width': 30,
               'use_metadata': True,
               'use_features': True,
               'model_heads': True,
               'coref_depth': 2,
               'lm_path': None,
               'lm_model_name': None,
               'lm_layers': 1,
               'lm_size': 768,
               'coarse_to_fine': True,
               'max_gradient_norm': 5.0,
               'lstm_dropout_rate': 0.4,
               'lexical_dropout_rate': 0.5,
               'dropout_rate': 0.2,
               'optimizer': 'adam',
               'learning_rate': 0.001,
               'decay_rate': 0.999,
               'decay_frequency': 100,
               'genres': ['all'],
               'eval_frequency': 5000,
               'report_frequency': 100,
               'use_gold': False,
               'log_root': 'logs',
               'log_dir': 'logs/best',
               'datapath': '.'
               }


def test_tensorize_example():
    example = {
        'doc_key': 'test',
        'sentences': [['Een', 'zin', '.']],
        'clusters': []
    }
    with tf.Session() as session:
        model = e2edutch.coref_model.CorefModel(test_config)
        tensors = model.tensorize_example(example, False)
        (tokens, context_word_emb, head_word_emb, lm_emb,
         char_index, text_len, genre, is_training,
         gold_starts, gold_ends, cluster_ids) = tensors
    tf.reset_default_graph()
    assert tokens.shape == (1, 3)
    assert context_word_emb.shape == (1, 3, 10)
    assert head_word_emb.shape == (1, 3, 10)
    assert lm_emb.shape == (0, 0,
                            test_config['lm_size'], test_config['lm_layers'])
    assert len(gold_starts) == 0



import e2edutch.bert

def test_encode_sentences():
    sentences = ['Dit is een zin .'.split(' '), 'Het is slechts een voorbeeld .'.split(' ')]
    tokenizer, model = e2edutch.bert.load_bert('bertje')
    emb = e2edutch.bert.encode_sentences(sentences, tokenizer, model)
    assert emb.shape == (2, 6, 768, 1)



import re
import os
import sys
import json
import collections
import logging
import argparse

from e2edutch import util
from e2edutch import conll

logger = logging.getLogger('e2edutch')


class DocumentState(object):
    def __init__(self):
        self.doc_key = None
        self.text = []
        self.sentences = []
        self.constituents = {}
        self.const_stack = []
        self.ner = {}
        self.ner_stack = []
        self.clusters = collections.defaultdict(list)
        self.coref_stacks = collections.defaultdict(list)

    def assert_empty(self):
        assert self.doc_key is None
        assert len(self.text) == 0
        assert len(self.sentences) == 0
        assert len(self.constituents) == 0
        assert len(self.const_stack) == 0
        assert len(self.ner) == 0
        assert len(self.ner_stack) == 0
        assert len(self.coref_stacks) == 0
        assert len(self.clusters) == 0

    def assert_finalizable(self):
        assert self.doc_key is not None
        assert len(self.text) == 0
        assert len(self.sentences) > 0
        assert len(self.const_stack) == 0
        assert len(self.ner_stack) == 0
        assert all(len(s) == 0 for s in self.coref_stacks.values())

    def span_dict_to_list(self, span_dict):
        return [(s, e, l) for (s, e), l in span_dict.items()]

    def finalize(self):
        merged_clusters = []
        for c1 in self.clusters.values():
            existing = None
            for m in c1:
                for c2 in merged_clusters:
                    if m in c2:
                        existing = c2
                        break
                if existing is not None:
                    break
            if existing is not None:
                print("Merging clusters (shouldn't happen very often.)")
                print(self.doc_key, m)
                existing.update(c1)
            else:
                merged_clusters.append(set(c1))
        merged_clusters = [list(c) for c in merged_clusters]
        all_mentions = util.flatten(merged_clusters)
        assert len(all_mentions) == len(set(all_mentions))

        return {
            "doc_key": self.doc_key,
            "sentences": self.sentences,
            "clusters": merged_clusters
        }


def normalize_word(word):
    if word == "/." or word == "/?":
        return word[1:]
    else:
        return word


def handle_bit(word_index, bit, stack, spans):
    asterisk_idx = bit.find("*")
    if asterisk_idx >= 0:
        open_parens = bit[:asterisk_idx]
        close_parens = bit[asterisk_idx + 1:]
    else:
        open_parens = bit[:-1]
        close_parens = bit[-1]

    current_idx = open_parens.find("(")
    while current_idx >= 0:
        next_idx = open_parens.find("(", current_idx + 1)
        if next_idx >= 0:
            label = open_parens[current_idx + 1:next_idx]
        else:
            label = open_parens[current_idx + 1:]
        stack.append((word_index, label))
        current_idx = next_idx

    for c in close_parens:
        assert c == ")"
        open_index, label = stack.pop()
        current_span = (open_index, word_index)
        """
    if current_span in spans:
      spans[current_span] += "_" + label
    else:
      spans[current_span] = label
    """
        spans[current_span] = label


def handle_line(line, document_state, labels, stats, word_col):
    begin_document_match = re.match(conll.BEGIN_DOCUMENT_REGEX, line)
    if begin_document_match:
        document_state.assert_empty()
        document_state.doc_key = conll.get_doc_key(
            *begin_document_match.groups())
        return None
    elif line.startswith("#end document"):
        if len(document_state.text) > 0:  # no newline before end document
            stats["max_sent_len"] = max(
                len(document_state.text), stats["max_sent_len"])
            stats["num_sents"] += 1
            document_state.sentences.append(tuple(document_state.text))
            del document_state.text[:]
        document_state.assert_finalizable()
        finalized_state = document_state.finalize()
        stats["num_clusters"] += len(finalized_state["clusters"])
        stats["num_mentions"] += sum(len(c)
                                     for c in finalized_state["clusters"])
        # labels["const_labels"].update(
        #     l for _, _, l in finalized_state["constituents"])
        # labels["ner"].update(l for _, _, l in finalized_state["ner"])
        return finalized_state
    else:
        row = line.split()
        if len(row) == 0 and len(document_state.text) > 0:
            stats["max_sent_len"] = max(
                len(document_state.text), stats["max_sent_len"])
            stats["num_sents"] += 1
            document_state.sentences.append(tuple(document_state.text))
            del document_state.text[:]
            return None
        elif len(row) == 0 and len(document_state.text) == 0:
            return None
        assert len(row) >= 4

        word = normalize_word(row[word_col])
        coref = row[-1]

        word_index = (len(document_state.text)
                      + sum(len(s) for s in document_state.sentences))
        document_state.text.append(word)

        if coref != "-" and coref != '_':
            for segment in coref.split("|"):
                if segment[0] == "(":
                    if segment[-1] == ")":
                        cluster_id = int(segment[1:-1])
                        document_state.clusters[cluster_id].append(
                            (word_index, word_index))
                    else:
                        cluster_id = int(segment[1:])
                        document_state.coref_stacks[cluster_id].append(
                            word_index)
                elif segment[-1] == ")":
                    cluster_id = int(segment[:-1])
                    start = document_state.coref_stacks[cluster_id].pop()
                    document_state.clusters[cluster_id].append(
                        (start, word_index))
        return None


def minimize_partition(input_path, labels, stats, word_col):
    with open(input_path, "r") as input_file:
        document_state = DocumentState()
        for line in input_file.readlines():
            document = handle_line(line, document_state,
                                   labels, stats, word_col)
            if document is not None:
                yield document
                document_state = DocumentState()


def minimize_partition_file(
        input_path, labels, stats, word_col, output_file=None):
    if output_file is None:
        output_path = "{}.jsonlines".format(os.path.splitext(input_path)[0])
        output_file = open(output_path, "w")
    count = 0
    logger.info("Minimizing {}".format(input_path))
    for document in minimize_partition(input_path, labels, stats, word_col):
        output_file.write(json.dumps(document))
        output_file.write("\n")
        count += 1
    logger.info("Wrote {} documents to {}".format(count, output_path))


def get_parser():
    parser = argparse.ArgumentParser()
    parser.add_argument('input_filename')
    parser.add_argument('-o', '--output_file',
                        type=argparse.FileType('w'), default=sys.stdout)
    parser.add_argument('-v', '--verbose', action='store_true')
    return parser


if __name__ == "__main__":
    parser = get_parser()
    args = parser.parse_args()
    input_path = args.input_filename
    output_file = args.output_file
    word_col = 2
    labels = collections.defaultdict(set)
    stats = collections.defaultdict(int)
    minimize_partition_file(input_path, labels, stats, word_col, output_file)
    for k, v in labels.items():
        print("{} = [{}]".format(k, ", ".join(
            "\"{}\"".format(label) for label in v)))
    for k, v in stats.items():
        print("{} = {}".format(k, v))



import logging
import os
import random
import math
import json
import threading
import numpy as np
import h5py
import tensorflow.compat.v1 as tf

from . import bert
from . import util
from . import coref_ops
from . import conll
from . import metrics

tf.disable_v2_behavior()

logger = logging.getLogger('e2edutch')

# When running as a Stanza Processor, we will be instantiating the CorefModel
# for each call to the pipeline, making interactive use impossible
# To speed things up, cache the slowest part here.
e2e_cached_embedding = None
e2e_cached_bert_model = None
e2e_cached_bert_tokenizer = None


class CorefModel(object):
    def __init__(self, config):
        # these are variables in the outer scope
        global e2e_cached_embedding
        global e2e_cached_bert_tokenizer
        global e2e_cached_bert_model

        self.config = config

        logger.info("Loading context embeddings..")
        self.context_embeddings = util.EmbeddingDictionary(
            config["context_embeddings"], config['datapath'],
            maybe_cache=e2e_cached_embedding)

        # cache the embeddings for reuse
        e2e_cached_embedding = self.context_embeddings

        logger.info("Loading head embeddings..")
        self.head_embeddings = util.EmbeddingDictionary(
            config["context_embeddings"], config['datapath'],
            maybe_cache=e2e_cached_embedding)

        self.char_embedding_size = config["char_embedding_size"]
        self.char_dict = util.load_char_dict(
            os.path.join(config['datapath'],
                         config["char_vocab_path"]))

        self.max_span_width = config["max_span_width"]
        self.genres = {g: i for i, g in enumerate(config["genres"])}
        if config["lm_path"]:
            self.lm_file = h5py.File(os.path.join(config['datapath'],
                                                  self.config["lm_path"]), "r")
        else:
            self.lm_file = None

        if config["lm_model_name"]:
            logger.info("Loading BERT model...")
            if e2e_cached_bert_model and e2e_cached_bert_tokenizer:
                # reuse cached version
                self.bert_tokenizer = e2e_cached_bert_tokenizer
                self.bert_model = e2e_cached_bert_model
            else:
                # load the model...
                self.bert_tokenizer, self.bert_model = bert.load_bert(
                    self.config["lm_model_name"])

                # ...and cache for next time
                e2e_cached_bert_tokenizer = self.bert_tokenizer
                e2e_cached_bert_model = self.bert_model
        else:
            self.bert_tokenizer = None
            self.bert_model = None
        self.lm_layers = self.config["lm_layers"]
        self.lm_size = self.config["lm_size"]
        self.eval_data = None  # Load eval data lazily.

        input_props = []
        input_props.append((tf.string, [None, None]))  # Tokens.

        # Context embeddings.
        input_props.append(
            (tf.float32, [None, None, self.context_embeddings.size]))

        # Head embeddings.
        input_props.append(
            (tf.float32, [None, None, self.head_embeddings.size]))

        # LM embeddings.
        input_props.append(
            (tf.float32, [None, None, self.lm_size, self.lm_layers]))

        # Character indices.
        input_props.append((tf.int32, [None, None, None]))
        input_props.append((tf.int32, [None]))  # Text lengths..
        input_props.append((tf.int32, []))  # Genre.
        input_props.append((tf.bool, []))  # Is training.
        input_props.append((tf.int32, [None]))  # Gold starts.
        input_props.append((tf.int32, [None]))  # Gold ends.
        input_props.append((tf.int32, [None]))  # Cluster ids.

        self.queue_input_tensors = [tf.placeholder(
            dtype, shape) for dtype, shape in input_props]
        dtypes, shapes = zip(*input_props)
        queue = tf.PaddingFIFOQueue(capacity=10, dtypes=dtypes, shapes=shapes)
        self.enqueue_op = queue.enqueue(self.queue_input_tensors)
        self.input_tensors = queue.dequeue()

        self.predictions, self.loss = self.get_predictions_and_loss(
            *self.input_tensors)
        self.global_step = tf.Variable(0, name="global_step", trainable=False)
        self.reset_global_step = tf.assign(self.global_step, 0)
        learning_rate = tf.train.exponential_decay(self.config["learning_rate"],
                                                   self.global_step,
                                                   self.config["decay_frequency"],
                                                   self.config["decay_rate"],
                                                   staircase=True)
        trainable_params = tf.trainable_variables()
        gradients = tf.gradients(self.loss, trainable_params)
        gradients, _ = tf.clip_by_global_norm(
            gradients, self.config["max_gradient_norm"])
        optimizers = {
            "adam": tf.train.AdamOptimizer,
            "sgd": tf.train.GradientDescentOptimizer
        }
        optimizer = optimizers[self.config["optimizer"]](learning_rate)
        self.train_op = optimizer.apply_gradients(
            zip(gradients, trainable_params), global_step=self.global_step)

    def start_enqueue_thread(self, session):
        with open(self.config["train_path"]) as f:
            train_examples = [json.loads(jsonline)
                              for jsonline in f.readlines()]

        def _enqueue_loop():
            while True:
                random.shuffle(train_examples)
                for example in train_examples:
                    tensorized_example = self.tensorize_example(
                        example, is_training=True)
                    feed_dict = dict(
                        zip(self.queue_input_tensors, tensorized_example))
                    session.run(self.enqueue_op, feed_dict=feed_dict)
        enqueue_thread = threading.Thread(target=_enqueue_loop)
        enqueue_thread.daemon = True
        enqueue_thread.start()

    def restore(self, session):
        # Don't try to restore unused variables from the TF-Hub ELMo module.
        vars_to_restore = [
            v for v in tf.global_variables() if "module/" not in v.name]
        saver = tf.train.Saver(vars_to_restore)
        checkpoint_path = os.path.join(
            self.config['log_root'], self.config['log_dir'], "model.max.ckpt")

        logger.info("Restoring coref model from {}".format(checkpoint_path))
        session.run(tf.global_variables_initializer())
        saver.restore(session, checkpoint_path)

    def load_lm_embeddings(self, doc_key, sentences):
        if self.lm_file is None and self.bert_model is None:
            # No LM specified
            return np.zeros([0, 0, self.lm_size, self.lm_layers])
        elif self.lm_file is None:
            # No cache file, encode on the fly
            lm_emb = bert.encode_sentences(
                sentences, self.bert_tokenizer, self.bert_model)
            return lm_emb
        file_key = doc_key.replace("/", ":")
        group = self.lm_file.get(file_key, None)
        if group is None and self.bert_model is not None:
            # Document not cached, encode on the fly
            lm_emb = bert.encode_sentences(
                sentences, self.bert_tokenizer, self.bert_model)
        elif group is not None:
            # Load encoding from cache file
            num_sentences = len(list(group.keys()))
            sentences = [group[str(i)][...] for i in range(num_sentences)]
            lm_emb = np.zeros([num_sentences, max(s.shape[0]
                                                  for s in sentences),
                               self.lm_size, self.lm_layers])
            for i, s in enumerate(sentences):
                lm_emb[i, :s.shape[0], :, :] = s
        return lm_emb

    def tensorize_mentions(self, mentions):
        if len(mentions) > 0:
            starts, ends = zip(*mentions)
        else:
            starts, ends = [], []
        return np.array(starts), np.array(ends)

    def tensorize_span_labels(self, tuples, label_dict):
        if len(tuples) > 0:
            starts, ends, labels = zip(*tuples)
        else:
            starts, ends, labels = [], [], []
        return (np.array(starts),
                np.array(ends),
                np.array([label_dict[c] for c in labels]))

    def tensorize_example(self, example, is_training):
        clusters = example["clusters"] if "clusters" in example else []

        gold_mentions = sorted(tuple(m) for m in util.flatten(clusters))
        gold_mention_map = {m: i for i, m in enumerate(gold_mentions)}
        cluster_ids = np.zeros(len(gold_mentions))
        for cluster_id, cluster in enumerate(clusters):
            for mention in cluster:
                cluster_ids[gold_mention_map[tuple(mention)]] = cluster_id + 1

        sentences = example["sentences"]

        max_sentence_length = max(len(s) for s in sentences)
        max_word_length = max(max(max(len(w) for w in s)
                                  for s in sentences),
                              max(self.config["filter_widths"]))
        text_len = np.array([len(s) for s in sentences])
        tokens = [[""] * max_sentence_length for _ in sentences]
        context_word_emb = np.zeros(
            [len(sentences),
             max_sentence_length,
             self.context_embeddings.size])
        head_word_emb = np.zeros(
            [len(sentences),
             max_sentence_length,
             self.head_embeddings.size])
        char_index = np.zeros(
            [len(sentences), max_sentence_length, max_word_length])
        for i, sentence in enumerate(sentences):
            for j, word in enumerate(sentence):
                tokens[i][j] = word
                context_word_emb[i, j] = self.context_embeddings[word]
                head_word_emb[i, j] = self.head_embeddings[word]
                char_index[i, j, :len(word)] = [self.char_dict[c]
                                                for c in word]
        tokens = np.array(tokens)

        genre = self.genres[example.get('genre', 'all')]

        gold_starts, gold_ends = self.tensorize_mentions(gold_mentions)

        lm_emb = self.load_lm_embeddings(
            example["doc_key"], example["sentences"])

        example_tensors = (tokens, context_word_emb, head_word_emb, lm_emb,
                           char_index, text_len, genre, is_training,
                           gold_starts, gold_ends, cluster_ids)

        if is_training and len(
                sentences) > self.config["max_training_sentences"]:
            return self.truncate_example(*example_tensors)
        else:
            return example_tensors

    def truncate_example(self, tokens, context_word_emb, head_word_emb, lm_emb,
                         char_index, text_len, genre, is_training, gold_starts,
                         gold_ends, cluster_ids):
        max_training_sentences = self.config["max_training_sentences"]
        num_sentences = context_word_emb.shape[0]
        assert num_sentences > max_training_sentences

        sentence_offset = random.randint(
            0, num_sentences - max_training_sentences)
        word_offset = text_len[:sentence_offset].sum()
        num_words = text_len[sentence_offset:sentence_offset
                             + max_training_sentences].sum()
        tokens = tokens[sentence_offset:sentence_offset
                        + max_training_sentences, :]
        context_word_emb = context_word_emb[sentence_offset:sentence_offset
                                            + max_training_sentences, :, :]
        head_word_emb = head_word_emb[sentence_offset:sentence_offset
                                      + max_training_sentences, :, :]
        lm_emb = lm_emb[sentence_offset:sentence_offset
                        + max_training_sentences, :, :, :]
        char_index = char_index[sentence_offset:sentence_offset
                                + max_training_sentences, :, :]
        text_len = text_len[sentence_offset:sentence_offset
                            + max_training_sentences]

        gold_spans = np.logical_and(
            gold_ends >= word_offset, gold_starts < word_offset + num_words)
        gold_starts = gold_starts[gold_spans] - word_offset
        gold_ends = gold_ends[gold_spans] - word_offset
        cluster_ids = cluster_ids[gold_spans]

        return (tokens, context_word_emb, head_word_emb, lm_emb, char_index,
                text_len, genre, is_training, gold_starts, gold_ends,
                cluster_ids)

    def get_candidate_labels(self, candidate_starts, candidate_ends,
                             labeled_starts, labeled_ends, labels):
        # [num_labeled, num_candidates]
        same_start = tf.equal(tf.expand_dims(labeled_starts, 1),
                              tf.expand_dims(candidate_starts, 0))
        # [num_labeled, num_candidates]
        same_end = tf.equal(tf.expand_dims(labeled_ends, 1),
                            tf.expand_dims(candidate_ends, 0))
        # [num_labeled, num_candidates]
        same_span = tf.logical_and(same_start, same_end)
        candidate_labels = tf.matmul(tf.expand_dims(
            labels, 0), tf.to_int32(same_span))  # [1, num_candidates]
        candidate_labels = tf.squeeze(candidate_labels, 0)  # [num_candidates]
        return candidate_labels

    def get_dropout(self, dropout_rate, is_training):
        return 1 - (tf.to_float(is_training) * dropout_rate)

    def coarse_to_fine_pruning(self, top_span_emb, top_span_mention_scores, c):
        k = util.shape(top_span_emb, 0)
        top_span_range = tf.range(k)  # [k]
        antecedent_offsets = tf.expand_dims(
            top_span_range, 1) - tf.expand_dims(top_span_range, 0)  # [k, k]
        antecedents_mask = antecedent_offsets >= 1  # [k, k]
        # [k, k]
        fast_antecedent_scores = (tf.expand_dims(top_span_mention_scores, 1)
                                  + tf.expand_dims(top_span_mention_scores, 0))
        # [k, k]
        fast_antecedent_scores += tf.log(tf.to_float(antecedents_mask))
        # [k, k]
        fast_antecedent_scores += self.get_fast_antecedent_scores(top_span_emb)

        _, top_antecedents = tf.nn.top_k(
            fast_antecedent_scores, c, sorted=False)  # [k, c]
        top_antecedents_mask = util.batch_gather(
            antecedents_mask, top_antecedents)  # [k, c]
        top_fast_antecedent_scores = util.batch_gather(
            fast_antecedent_scores, top_antecedents)  # [k, c]
        top_antecedent_offsets = util.batch_gather(
            antecedent_offsets, top_antecedents)  # [k, c]
        return (top_antecedents, top_antecedents_mask,
                top_fast_antecedent_scores, top_antecedent_offsets)

    def distance_pruning(self, top_span_emb, top_span_mention_scores, c):
        k = util.shape(top_span_emb, 0)
        top_antecedent_offsets = tf.tile(
            tf.expand_dims(tf.range(c) + 1, 0), [k, 1])  # [k, c]
        raw_top_antecedents = tf.expand_dims(
            tf.range(k), 1) - top_antecedent_offsets  # [k, c]
        top_antecedents_mask = raw_top_antecedents >= 0  # [k, c]
        top_antecedents = tf.maximum(raw_top_antecedents, 0)  # [k, c]

        # [k, c]
        top_fast_antecedent_scores = tf.expand_dims(
            top_span_mention_scores, 1) + tf.gather(top_span_mention_scores,
                                                    top_antecedents)
        # [k, c]
        top_fast_antecedent_scores += tf.log(tf.to_float(top_antecedents_mask))
        return (top_antecedents, top_antecedents_mask,
                top_fast_antecedent_scores, top_antecedent_offsets)

    def get_predictions_and_loss(self, tokens, context_word_emb, head_word_emb,
                                 lm_emb, char_index, text_len, genre,
                                 is_training, gold_starts, gold_ends,
                                 cluster_ids):
        self.dropout = self.get_dropout(
            self.config["dropout_rate"], is_training)
        self.lexical_dropout = self.get_dropout(
            self.config["lexical_dropout_rate"], is_training)
        self.lstm_dropout = self.get_dropout(
            self.config["lstm_dropout_rate"], is_training)

        num_sentences = tf.shape(context_word_emb)[0]
        max_sentence_length = tf.shape(context_word_emb)[1]

        context_emb_list = [context_word_emb]
        head_emb_list = [head_word_emb]

        if self.config["char_embedding_size"] > 0:
            # [num_sentences, max_sentence_length, max_word_length, emb]
            char_emb = tf.gather(tf.get_variable("char_embeddings", [len(
                self.char_dict), self.config["char_embedding_size"]]),
                char_index)

            # [num_sentences * max_sentence_length, max_word_length, emb]
            flattened_char_emb = tf.reshape(char_emb,
                                            [num_sentences
                                             * max_sentence_length,
                                             util.shape(char_emb, 2),
                                             util.shape(char_emb, 3)])

            # [num_sentences * max_sentence_length, emb]
            flattened_aggregated_char_emb = util.cnn(
                flattened_char_emb, self.config["filter_widths"],
                self.config["filter_size"])

            # [num_sentences, max_sentence_length, emb]
            aggregated_char_emb = tf.reshape(flattened_aggregated_char_emb,
                                             [num_sentences,
                                              max_sentence_length,
                                              util.shape(
                                                  flattened_aggregated_char_emb,
                                                  1)])
            context_emb_list.append(aggregated_char_emb)
            head_emb_list.append(aggregated_char_emb)

        lm_emb_size = util.shape(lm_emb, 2)
        lm_num_layers = util.shape(lm_emb, 3)
        with tf.variable_scope("lm_aggregation"):
            self.lm_weights = tf.nn.softmax(tf.get_variable(
                "lm_scores", [lm_num_layers],
                initializer=tf.constant_initializer(0.0)))
            self.lm_scaling = tf.get_variable(
                "lm_scaling", [],
                initializer=tf.constant_initializer(1.0))
        flattened_lm_emb = tf.reshape(
            lm_emb, [num_sentences
                     * max_sentence_length
                     * lm_emb_size,
                     lm_num_layers])

        # [num_sentences * max_sentence_length * emb, 1]
        flattened_aggregated_lm_emb = tf.matmul(flattened_lm_emb,
                                                tf.expand_dims(
                                                    self.lm_weights, 1))
        aggregated_lm_emb = tf.reshape(flattened_aggregated_lm_emb, [
                                       num_sentences, max_sentence_length,
                                       lm_emb_size])
        aggregated_lm_emb *= self.lm_scaling
        context_emb_list.append(aggregated_lm_emb)

        # [num_sentences, max_sentence_length, emb]
        context_emb = tf.concat(context_emb_list, 2)
        # [num_sentences, max_sentence_length, emb]
        head_emb = tf.concat(head_emb_list, 2)
        # [num_sentences, max_sentence_length, emb]
        context_emb = tf.nn.dropout(context_emb, self.lexical_dropout)
        # [num_sentences, max_sentence_length, emb]
        head_emb = tf.nn.dropout(head_emb, self.lexical_dropout)

        # [num_sentence, max_sentence_length]
        text_len_mask = tf.sequence_mask(text_len, maxlen=max_sentence_length)

        context_outputs = self.lstm_contextualize(
            context_emb, text_len, text_len_mask)  # [num_words, emb]
        num_words = util.shape(context_outputs, 0)

        genre_emb = tf.gather(tf.get_variable("genre_embeddings", [len(
            self.genres), self.config["feature_size"]]), genre)  # [emb]

        # [num_sentences, max_sentence_length]
        sentence_indices = tf.tile(tf.expand_dims(
            tf.range(num_sentences), 1), [1, max_sentence_length])
        flattened_sentence_indices = self.flatten_emb_by_sentence(
            sentence_indices, text_len_mask)  # [num_words]
        flattened_head_emb = self.flatten_emb_by_sentence(
            head_emb, text_len_mask)  # [num_words]
        candidate_starts = tf.tile(tf.expand_dims(tf.range(num_words), 1), [
            1, self.max_span_width])  # [num_words, max_span_width]
        candidate_ends = candidate_starts + \
            tf.expand_dims(tf.range(self.max_span_width),
                           0)  # [num_words, max_span_width]
        # [num_words, max_span_width]
        candidate_start_sentence_indices = tf.gather(
            flattened_sentence_indices, candidate_starts)
        # [num_words, max_span_width]
        candidate_end_sentence_indices = tf.gather(flattened_sentence_indices,
                                                   tf.minimum(
                                                       candidate_ends,
                                                       num_words - 1))
        # [num_words, max_span_width]
        candidate_mask = tf.logical_and(candidate_ends < num_words, tf.equal(
            candidate_start_sentence_indices, candidate_end_sentence_indices))
        flattened_candidate_mask = tf.reshape(
            candidate_mask, [-1])  # [num_words * max_span_width]
        # [num_candidates]
        candidate_starts = tf.boolean_mask(tf.reshape(
            candidate_starts, [-1]), flattened_candidate_mask)
        # [num_candidates]
        candidate_ends = tf.boolean_mask(tf.reshape(
            candidate_ends, [-1]), flattened_candidate_mask)
        # [num_candidates]
        candidate_sentence_indices = tf.boolean_mask(tf.reshape(
            candidate_start_sentence_indices, [-1]), flattened_candidate_mask)
        # [num_candidates]
        candidate_cluster_ids = self.get_candidate_labels(candidate_starts,
                                                          candidate_ends,
                                                          gold_starts,
                                                          gold_ends,
                                                          cluster_ids)
        # [num_candidates, emb]
        candidate_span_emb = self.get_span_emb(flattened_head_emb,
                                               context_outputs,
                                               candidate_starts,
                                               candidate_ends)
        candidate_mention_scores = self.get_mention_scores(
            candidate_span_emb)  # [k, 1]
        candidate_mention_scores = tf.squeeze(
            candidate_mention_scores, 1)  # [k]

        if self.config['use_gold']:
            candidates_spans = tf.stack(
                [candidate_starts, candidate_ends], axis=1)
            gold_spans = tf.stack([gold_starts, gold_ends], axis=1)
            same_span = tf.equal(tf.expand_dims(
                gold_spans, 1), tf.expand_dims(candidates_spans, 0))
            top_span_indices = tf.reduce_any(
                tf.reduce_all(same_span, axis=2), axis=0)
            top_span_indices = tf.squeeze(tf.where(top_span_indices), axis=1)
            k = tf.cast(util.shape(top_span_indices, 0), tf.int32)

        else:
            k = tf.to_int32(tf.floor(tf.to_float(
                tf.shape(context_outputs)[0]) * self.config["top_span_ratio"]))
            top_span_indices = coref_ops.extract_spans(
                tf.expand_dims(candidate_mention_scores, 0),
                tf.expand_dims(candidate_starts, 0),
                tf.expand_dims(candidate_ends, 0),
                tf.expand_dims(k, 0),
                util.shape(context_outputs, 0),
                True)  # [1, k]
            top_span_indices.set_shape([1, None])
            top_span_indices = tf.squeeze(top_span_indices, 0)  # [k]

        top_span_starts = tf.gather(candidate_starts, top_span_indices)  # [k]
        top_span_ends = tf.gather(candidate_ends, top_span_indices)  # [k]
        top_span_emb = tf.gather(
            candidate_span_emb, top_span_indices)  # [k, emb]

        top_span_cluster_ids = tf.gather(
            candidate_cluster_ids, top_span_indices)  # [k]
        top_span_mention_scores = tf.gather(
            candidate_mention_scores, top_span_indices)  # [k]
        top_span_sentence_indices = tf.gather(
            candidate_sentence_indices, top_span_indices)  # [k]

        c = tf.minimum(self.config["max_top_antecedents"], k)

        if self.config["coarse_to_fine"]:
            (top_antecedents,
             top_antecedents_mask,
             top_fast_antecedent_scores,
             top_antecedent_offsets) = self.coarse_to_fine_pruning(
                top_span_emb, top_span_mention_scores, c)
        else:
            (top_antecedents,
             top_antecedents_mask,
             top_fast_antecedent_scores,
             top_antecedent_offsets) = self.distance_pruning(
                top_span_emb, top_span_mention_scores, c)
        dummy_scores_nomention = tf.expand_dims(
            top_span_mention_scores * -1, 1)  # tf.zeros([k, 1])  # [k, 1]
        dummy_scores_first = tf.zeros([k, 1])  # [k, 1]
        for i in range(self.config["coref_depth"]):
            with tf.variable_scope("coref_layer", reuse=(i > 0)):
                top_antecedent_emb = tf.gather(
                    top_span_emb, top_antecedents)  # [k, c, emb]
                top_antecedent_scores = (top_fast_antecedent_scores
                                         + self.get_slow_antecedent_scores(
                                             top_span_emb,
                                             top_antecedents,
                                             top_antecedent_emb,
                                             top_antecedent_offsets,
                                             genre_emb))  # [k, c]
                top_antecedent_weights = tf.nn.softmax(
                    tf.concat([dummy_scores_nomention,
                               dummy_scores_first,
                               top_antecedent_scores], 1))  # [k, c + 2]
                top_antecedent_emb = tf.concat(
                    [tf.expand_dims(top_span_emb, 1),
                     tf.expand_dims(top_span_emb, 1),
                     top_antecedent_emb], 1)  # [k, c + 1, emb]
                # [k, emb]
                attended_span_emb = tf.reduce_sum(tf.expand_dims(
                    top_antecedent_weights, 2) * top_antecedent_emb, 1)
                with tf.variable_scope("f"):
                    f = tf.sigmoid(util.projection(
                        tf.concat([top_span_emb, attended_span_emb], 1),
                        util.shape(top_span_emb, -1)))  # [k, emb]
                    top_span_emb = f * attended_span_emb + \
                        (1 - f) * top_span_emb  # [k, emb]
        # [k, c + 2]
        top_antecedent_scores = tf.concat(
            [dummy_scores_nomention,
             dummy_scores_first,
             top_antecedent_scores], 1)

        top_antecedent_cluster_ids = tf.gather(
            top_span_cluster_ids, top_antecedents)  # [k, c]
        # [k, c]
        top_antecedent_cluster_ids += tf.to_int32(
            tf.log(tf.to_float(top_antecedents_mask)))
        same_cluster_indicator = tf.equal(
            top_antecedent_cluster_ids,
            tf.expand_dims(top_span_cluster_ids, 1))  # [k, c]
        non_dummy_indicator = tf.expand_dims(
            top_span_cluster_ids > 0, 1)  # [k, 1]
        pairwise_labels = tf.logical_and(
            same_cluster_indicator, non_dummy_indicator)  # [k, c]
        dummy_labels_nomention = tf.expand_dims(
            tf.equal(top_span_cluster_ids, 0), 1)  # [k, 1]
        dummy_labels_first = tf.logical_not(tf.reduce_any(
            tf.concat(
                [dummy_labels_nomention, pairwise_labels], 1), 1,
            keepdims=True))  # [k, 1]
        top_antecedent_labels = tf.concat(
            [dummy_labels_nomention, dummy_labels_first, pairwise_labels], 1)  # [k, c + 1]
        loss = self.softmax_loss(
            top_antecedent_scores, top_antecedent_labels)  # [k]
        loss = tf.reduce_sum(loss)  # []

        return [candidate_starts, candidate_ends, candidate_mention_scores,
                top_span_starts, top_span_ends, top_antecedents,
                top_antecedent_scores], loss

    def get_span_emb(self, head_emb, context_outputs, span_starts, span_ends):
        span_emb_list = []

        span_start_emb = tf.gather(context_outputs, span_starts)  # [k, emb]
        span_emb_list.append(span_start_emb)

        span_end_emb = tf.gather(context_outputs, span_ends)  # [k, emb]
        span_emb_list.append(span_end_emb)

        span_width = tf.minimum(
            1 + span_ends - span_starts, self.max_span_width)  # [k]

        if self.config["use_features"]:
            span_width_index = span_width - 1  # [k]
            span_width_emb = tf.gather(tf.get_variable("span_width_embeddings", [
                                       self.config["max_span_width"],
                                       self.config["feature_size"]]),
                                       span_width_index)  # [k, emb]
            span_width_emb = tf.nn.dropout(span_width_emb, self.dropout)
            span_emb_list.append(span_width_emb)

        if self.config["model_heads"]:
            # [k, max_span_width]
            span_indices = (tf.expand_dims(tf.range(
                self.config["max_span_width"]), 0)
                + tf.expand_dims(span_starts, 1))
            span_indices = tf.minimum(util.shape(
                context_outputs, 0) - 1, span_indices)  # [k, max_span_width]
            # [k, max_span_width, emb]
            span_text_emb = tf.gather(head_emb, span_indices)
            with tf.variable_scope("head_scores"):
                self.head_scores = util.projection(
                    context_outputs, 1)  # [num_words, 1]
            span_head_scores = tf.gather(
                self.head_scores, span_indices)  # [k, max_span_width, 1]
            span_mask = tf.expand_dims(tf.sequence_mask(
                span_width, self.config["max_span_width"],
                dtype=tf.float32), 2)  # [k, max_span_width, 1]
            span_head_scores += tf.log(span_mask)  # [k, max_span_width, 1]
            span_attention = tf.nn.softmax(
                span_head_scores, 1)  # [k, max_span_width, 1]
            span_head_emb = tf.reduce_sum(
                span_attention * span_text_emb, 1)  # [k, emb]
            span_emb_list.append(span_head_emb)

        span_emb = tf.concat(span_emb_list, 1)  # [k, emb]
        return span_emb  # [k, emb]

    def get_mention_scores(self, span_emb):
        with tf.variable_scope("mention_scores"):
            # [k, 1]
            return util.ffnn(
                span_emb, self.config["ffnn_depth"], self.config["ffnn_size"],
                1, self.dropout)

    def softmax_loss(self, antecedent_scores, antecedent_labels):
        gold_scores = antecedent_scores + \
            tf.log(tf.to_float(antecedent_labels))  # [k, max_ant + 1]
        marginalized_gold_scores = tf.reduce_logsumexp(gold_scores, [1])  # [k]
        log_norm = tf.reduce_logsumexp(antecedent_scores, [1])  # [k]
        return log_norm - marginalized_gold_scores  # [k]

    def bucket_distance(self, distances):
        """
        Places the given values (designed for distances) into
        10 semi-logscale buckets:
        [0, 1, 2, 3, 4, 5-7, 8-15, 16-31, 32-63, 64+].
        """
        logspace_idx = tf.to_int32(
            tf.floor(tf.log(tf.to_float(distances)) / math.log(2))) + 3
        use_identity = tf.to_int32(distances <= 4)
        combined_idx = use_identity * distances + \
            (1 - use_identity) * logspace_idx
        return tf.clip_by_value(combined_idx, 0, 9)

    def get_slow_antecedent_scores(
            self, top_span_emb, top_antecedents, top_antecedent_emb,
            top_antecedent_offsets, genre_emb):
        k = util.shape(top_span_emb, 0)
        c = util.shape(top_antecedents, 1)

        feature_emb_list = []

        if self.config["use_metadata"]:
            tiled_genre_emb = tf.tile(tf.expand_dims(
                tf.expand_dims(genre_emb, 0), 0), [k, c, 1])  # [k, c, emb]
            feature_emb_list.append(tiled_genre_emb)

        if self.config["use_features"]:
            antecedent_distance_buckets = self.bucket_distance(
                top_antecedent_offsets)  # [k, c]
            # [k, c]
            antecedent_distance_emb = tf.gather(tf.get_variable(
                                                "antecedent_distance_emb",
                                                [10,
                                                 self.config["feature_size"]]),
                                                antecedent_distance_buckets)
            feature_emb_list.append(antecedent_distance_emb)

        feature_emb = tf.concat(feature_emb_list, 2)  # [k, c, emb]
        feature_emb = tf.nn.dropout(feature_emb, self.dropout)  # [k, c, emb]

        target_emb = tf.expand_dims(top_span_emb, 1)  # [k, 1, emb]
        similarity_emb = top_antecedent_emb * target_emb  # [k, c, emb]
        target_emb = tf.tile(target_emb, [1, c, 1])  # [k, c, emb]

        # [k, c, emb]
        pair_emb = tf.concat(
            [target_emb, top_antecedent_emb, similarity_emb, feature_emb], 2)

        with tf.variable_scope("slow_antecedent_scores"):
            slow_antecedent_scores = util.ffnn(
                pair_emb, self.config["ffnn_depth"], self.config["ffnn_size"],
                1, self.dropout)  # [k, c, 1]
        slow_antecedent_scores = tf.squeeze(
            slow_antecedent_scores, 2)  # [k, c]
        return slow_antecedent_scores  # [k, c]

    def get_fast_antecedent_scores(self, top_span_emb):
        with tf.variable_scope("src_projection"):
            # [k, emb]
            source_top_span_emb = tf.nn.dropout(util.projection(
                top_span_emb, util.shape(top_span_emb, -1)), self.dropout)
        target_top_span_emb = tf.nn.dropout(
            top_span_emb, self.dropout)  # [k, emb]
        # [k, k]
        return tf.matmul(source_top_span_emb,
                         target_top_span_emb, transpose_b=True)

    def flatten_emb_by_sentence(self, emb, text_len_mask):
        num_sentences = tf.shape(emb)[0]
        max_sentence_length = tf.shape(emb)[1]

        emb_rank = len(emb.get_shape())
        if emb_rank == 2:
            flattened_emb = tf.reshape(
                emb, [num_sentences * max_sentence_length])
        elif emb_rank == 3:
            flattened_emb = tf.reshape(
                emb, [num_sentences * max_sentence_length, util.shape(emb, 2)])
        else:
            raise ValueError("Unsupported rank: {}".format(emb_rank))
        return tf.boolean_mask(flattened_emb, tf.reshape(
            text_len_mask, [num_sentences * max_sentence_length]))

    def lstm_contextualize(self, text_emb, text_len, text_len_mask):
        num_sentences = tf.shape(text_emb)[0]

        current_inputs = text_emb  # [num_sentences, max_sentence_length, emb]

        for layer in range(self.config["contextualization_layers"]):
            with tf.variable_scope("layer_{}".format(layer)):
                with tf.variable_scope("fw_cell"):
                    cell_fw = util.CustomLSTMCell(
                        self.config["contextualization_size"],
                        num_sentences,
                        self.lstm_dropout)
                with tf.variable_scope("bw_cell"):
                    cell_bw = util.CustomLSTMCell(
                        self.config["contextualization_size"],
                        num_sentences,
                        self.lstm_dropout)
                state_fw = tf.nn.rnn_cell.LSTMStateTuple(
                    tf.tile(cell_fw.initial_state.c, [
                        num_sentences, 1]),
                    tf.tile(cell_fw.initial_state.h, [
                        num_sentences, 1]))
                state_bw = tf.nn.rnn_cell.LSTMStateTuple(
                    tf.tile(cell_bw.initial_state.c, [
                        num_sentences, 1]),
                    tf.tile(cell_bw.initial_state.h, [
                        num_sentences, 1]))

                (fw_outputs, bw_outputs), _ = tf.nn.bidirectional_dynamic_rnn(
                    cell_fw=cell_fw,
                    cell_bw=cell_bw,
                    inputs=current_inputs,
                    sequence_length=text_len,
                    initial_state_fw=state_fw,
                    initial_state_bw=state_bw)

                # [num_sentences, max_sentence_length, emb]
                text_outputs = tf.concat([fw_outputs, bw_outputs], 2)
                text_outputs = tf.nn.dropout(text_outputs, self.lstm_dropout)
                if layer > 0:
                    # [num_sentences, max_sentence_length, emb]
                    highway_gates = tf.sigmoid(
                        util.projection(text_outputs,
                                        util.shape(text_outputs, 2)))
                    text_outputs = highway_gates * text_outputs + \
                        (1 - highway_gates) * current_inputs
                current_inputs = text_outputs

        return self.flatten_emb_by_sentence(text_outputs, text_len_mask)

    def get_predicted_antecedents(self, antecedents, antecedent_scores):
        predicted_antecedents = []
        for i, index in enumerate(np.argmax(antecedent_scores, axis=1) - 2):
            if index < 0:
                predicted_antecedents.append(index)
            else:
                predicted_antecedents.append(antecedents[i, index])
        return predicted_antecedents

    def get_predicted_clusters(
            self, top_span_starts, top_span_ends, predicted_antecedents):
        mention_to_predicted = {}
        predicted_clusters = []
        for i, predicted_index in enumerate(predicted_antecedents):
            mention = (int(top_span_starts[i]), int(top_span_ends[i]))
            if predicted_index == -1:  # First mention of cluster
                predicted_cluster = len(predicted_clusters)
                predicted_clusters.append([])
            elif predicted_index == -2:  # No mention, disregard
                continue
            else:
                assert i > predicted_index
                pred_antecedent = (int(top_span_starts[predicted_index]),
                                   int(top_span_ends[predicted_index]))
                if pred_antecedent in mention_to_predicted:
                    predicted_cluster = mention_to_predicted[pred_antecedent]
                else:  # Also add antecedent to cluster
                    predicted_cluster = len(predicted_clusters)
                    predicted_clusters.append([pred_antecedent])
                    mention_to_predicted[pred_antecedent] = predicted_cluster

            # add current mention
            predicted_clusters[predicted_cluster].append(mention)
            mention_to_predicted[mention] = predicted_cluster

        predicted_clusters = [tuple(pc) for pc in predicted_clusters]
        mention_to_predicted = {
            m: predicted_clusters[i] for m, i in mention_to_predicted.items()}

        return predicted_clusters, mention_to_predicted

    def evaluate_coref(self, top_span_starts, top_span_ends,
                       predicted_antecedents, gold_clusters, evaluator):
        gold_clusters = [tuple(tuple(m) for m in gc) for gc in gold_clusters]
        mention_to_gold = {}
        for gc in gold_clusters:
            for mention in gc:
                mention_to_gold[mention] = gc

        predicted_clusters, mention_to_predicted = self.get_predicted_clusters(
            top_span_starts, top_span_ends, predicted_antecedents)
        evaluator.update(predicted_clusters, gold_clusters,
                         mention_to_predicted, mention_to_gold)
        return predicted_clusters

    def load_eval_data(self):
        if self.eval_data is None:
            def load_line(line):
                example = json.loads(line)
                return self.tensorize_example(
                    example, is_training=False), example
            with open(self.config["eval_path"]) as f:
                self.eval_data = [load_line(l) for l in f.readlines()]
            logger.info("Loaded {} eval examples.".format(
                len(self.eval_data)))

    def evaluate(self, session, official_stdout=False):
        self.load_eval_data()

        coref_predictions = {}
        coref_evaluator = metrics.CorefEvaluator()

        for example_num, (tensorized_example,
                          example) in enumerate(self.eval_data):
            (_, _, _, _, _, _, _, _,
             gold_starts, gold_ends, _) = tensorized_example
            feed_dict = {i: t for i, t in zip(
                self.input_tensors, tensorized_example)}
            (candidate_starts, candidate_ends, candidate_mention_scores,
             top_span_starts, top_span_ends, top_antecedents,
             top_antecedent_scores) = session.run(
                self.predictions, feed_dict=feed_dict)
            predicted_antecedents = self.get_predicted_antecedents(
                top_antecedents, top_antecedent_scores)
            coref_predictions[example["doc_key"]] = self.evaluate_coref(
                top_span_starts, top_span_ends, predicted_antecedents,
                example["clusters"], coref_evaluator)
            if example_num % 10 == 0:
                logger.info("Evaluated {}/{} examples.".format(
                    example_num + 1, len(self.eval_data)))

        summary_dict = {}
        conll_results = conll.evaluate_conll(
            self.config["conll_eval_path"], coref_predictions, official_stdout)
        average_f1 = (sum(results["f"]
                          for results in conll_results.values())
                      / len(conll_results))
        summary_dict["Average F1 (conll)"] = average_f1
        logger.info("Average F1 (conll): {:.2f}%".format(average_f1))

        p, r, f = coref_evaluator.get_prf()
        summary_dict["Average F1 (py)"] = f
        logger.info("Average F1 (py): {:.2f}%".format(f * 100))
        summary_dict["Average precision (py)"] = p
        logger.info("Average precision (py): {:.2f}%".format(p * 100))
        summary_dict["Average recall (py)"] = r
        logger.info("Average recall (py): {:.2f}%".format(r * 100))

        return util.make_summary(summary_dict), average_f1



import sys
import json
import os
import io
import collections
import argparse
import logging

from e2edutch import conll
from e2edutch import minimize
from e2edutch import util
from e2edutch import coref_model as cm
from e2edutch import naf

import tensorflow.compat.v1 as tf

logger = logging.getLogger('e2edutch')


class Predictor(object):
    def __init__(self, model_name='final', config=None, verbose=False):
        if verbose:
            logger.setLevel(logging.INFO)

        if config:
            self.config = config
        else:
            # if no configuration is provided, try to get a default config.
            self.config = util.initialize_from_env(model_name=model_name)

        self.session = tf.compat.v1.Session()

        try:
            self.model = cm.CorefModel(self.config)
            self.model.restore(self.session)
        except ValueError:
            raise Exception("Trying to reload the model while the previous " +
                            "session hasn't been ended. Close the existing " +
                            "session with predictor.end_session()")

    def predict(self, example):
        """
        Predict coreference spans for a tokenized text.


        Args:
            example (dict): dict with the following fields:
                              sentences ([[str]])
                              doc_id (str)
                              clusters ([[(int, int)]]) (optional)

        Returns:
            [[(int, int)]]: a list of clusters. The items of the cluster are
                            spans, denoted by their start end end token index

        """
        tensorized_example = self.model.tensorize_example(
            example, is_training=False)
        feed_dict = {i: t for i, t in zip(
            self.model.input_tensors, tensorized_example)}
        _, _, _, top_span_starts, top_span_ends, top_antecedents, top_antecedent_scores = self.session.run(
            self.model.predictions, feed_dict=feed_dict)
        predicted_antecedents = self.model.get_predicted_antecedents(
            top_antecedents, top_antecedent_scores)
        predicted_clusters, _ = self.model.get_predicted_clusters(
            top_span_starts, top_span_ends, predicted_antecedents)

        return predicted_clusters

    def end_session(self):
        self.session.close()
        tf.reset_default_graph()


def get_parser():
    parser = argparse.ArgumentParser()
    parser.add_argument('config')
    parser.add_argument('input_filename')
    parser.add_argument('-o', '--output_file',
                        type=argparse.FileType('w'), default=sys.stdout)
    parser.add_argument('-f', '--format_out', default='conll',
                        choices=['conll', 'jsonlines', 'naf'])
    parser.add_argument('-c', '--word_col', type=int, default=2)
    parser.add_argument('--cfg_file',
                        type=str,
                        default=None,
                        help="config file")
    parser.add_argument('-v', '--verbose', action='store_true')
    return parser


def read_jsonlines(input_filename):
    for line in open(input_filename).readlines():
        example = json.loads(line)
        yield example


def main(args=None):
    parser = get_parser()
    args = parser.parse_args()
    if args.verbose:
        logger.setLevel(logging.INFO)

    # Input file in .jsonlines format or .conll.
    input_filename = args.input_filename

    ext_input = os.path.splitext(input_filename)[-1]
    if ext_input not in ['.conll', '.jsonlines', '.txt', '.naf']:
        raise Exception(
            'Input file should be .naf, .conll, .txt or .jsonlines, but is {}.'
            .format(ext_input))

    if ext_input == '.conll':
        labels = collections.defaultdict(set)
        stats = collections.defaultdict(int)
        docs = minimize.minimize_partition(
            input_filename, labels, stats, args.word_col)
    elif ext_input == '.jsonlines':
        docs = read_jsonlines(input_filename)
    elif ext_input == '.naf':
        naf_obj = naf.get_naf(input_filename)
        jsonlines_obj, term_ids, tok_ids = naf.get_jsonlines(naf_obj)
        docs = [jsonlines_obj]
    else:
        text = open(input_filename).read()
        docs = [util.create_example(text)]

    output_file = args.output_file

    config = util.initialize_from_env(cfg_file=args.cfg_file, model_cfg_file=args.config)
    predictor = Predictor(config=config)

    sentences = {}
    predictions = {}
    for example_num, example in enumerate(docs):
        example["predicted_clusters"], _ = predictor.predict(example)
        if args.format_out == 'jsonlines':
            output_file.write(json.dumps(example))
            output_file.write("\n")
        else:
            predictions[example['doc_key']] = example["predicted_clusters"]
            sentences[example['doc_key']] = example["sentences"]
        if example_num % 100 == 0:
            logger.info("Decoded {} examples.".format(example_num + 1))
    if args.format_out == 'conll':
        conll.output_conll(output_file, sentences, predictions)
    elif args.format_out == 'naf':
        # Check number of docs - what to do if multiple?
        # Create naf obj if input format was not naf
        if ext_input != '.naf':
            # To do: add linguistic processing layers for terms and tokens
            logger.warn(
                'Outputting NAF when input was not naf,'
                + 'no dependency information available')
            for doc_key in sentences:
                naf_obj, term_ids = naf.get_naf_from_sentences(
                    sentences[doc_key])
                naf_obj = naf.create_coref_layer(
                    naf_obj, predictions[doc_key], term_ids)
                naf_obj = naf.add_linguistic_processors(naf_obj)
                buffer = io.BytesIO()
                naf_obj.dump(buffer)
                output_file.write(buffer.getvalue().decode('utf-8'))
                # To do, make sepearate outputs?
                # TO do, use dependency information from conll?
        else:
            # We only have one input doc
            naf_obj = naf.create_coref_layer(
                naf_obj, example["predicted_clusters"], term_ids)
            naf_obj = naf.add_linguistic_processors(naf_obj)
            buffer = io.BytesIO()
            naf_obj.dump(buffer)
            output_file.write(buffer.getvalue().decode('utf-8'))


if __name__ == "__main__":
    main()



from tensorflow.python import pywrap_tensorflow
import pkg_resources

import tensorflow.compat.v1 as tf
tf.disable_v2_behavior()

coref_op_library = tf.load_op_library(
    pkg_resources.resource_filename("e2edutch", "coref_kernels.so"))

extract_spans = coref_op_library.extract_spans
tf.NotDifferentiable("ExtractSpans")



import e2edutch.metrics
import math


def test_f1():
    num = 10
    p_den = 100
    r_den = 50
    f1 = e2edutch.metrics.f1(num, p_den, num, r_den)
    assert math.isclose(f1, 2.0/15)


def test_f1_zero():
    num = 0
    p_den = 0
    r_den = 0
    f1 = e2edutch.metrics.f1(num, p_den, num, r_den)
    assert math.isclose(f1, 0)


def test_corefevaluator():
    evaluator = e2edutch.metrics.CorefEvaluator()
    assert evaluator.get_f1() == 0
    assert evaluator.get_recall() == 0
    assert evaluator.get_precision() == 0

    # empty update
    evaluator.update([], [], {}, {})
    assert evaluator.get_f1() == 0
    assert evaluator.get_recall() == 0
    assert evaluator.get_precision() == 0


def test_evaluate_documents():
    metrics = [e2edutch.metrics.muc,
               e2edutch.metrics.b_cubed,
               e2edutch.metrics.lea,
               e2edutch.metrics.ceafe]
    documents = []
    for metric in metrics:
        p, r, f = e2edutch.metrics.evaluate_documents(documents, metric)
        assert p == 0
        assert r == 0
        assert f == 0


def example_clusters():
    mentions = tuple([(i*2, i*2+1) for i in range(9)])
    a, b, c, d, e, f, g, h, i = mentions
    pred_clusters = [(a, b), (c, d), (f, g, h, i)]
    gold_clusters = [(a, b, c), (d, e, f, g)]
    mentions_to_gold = {m: cl for cl in tuple(gold_clusters) for m in cl}
    return pred_clusters, gold_clusters, mentions_to_gold


def test_muc():
    clusters = [((0, 1), (2, 3))]
    mentions_to_gold = {m: cl for cl in tuple(clusters) for m in cl}
    num, dem = e2edutch.metrics.muc(clusters, mentions_to_gold)
    assert math.isclose(num/dem, 1)

    clusters, _, mentions_to_gold = example_clusters()
    num, dem = e2edutch.metrics.muc(clusters, mentions_to_gold)
    assert math.isclose(num/dem, 0.4)


def test_b_cubed():
    clusters = [((0, 1), (2, 3))]
    mentions_to_gold = {m: cl for cl in tuple(clusters) for m in cl}
    num, dem = e2edutch.metrics.b_cubed(clusters, mentions_to_gold)
    assert math.isclose(num/dem, 1)

    clusters, _, mentions_to_gold = example_clusters()
    num, dem = e2edutch.metrics.b_cubed(clusters, mentions_to_gold)
    assert math.isclose(num/dem, 0.5)


def test_ceafe():
    clusters = [((0, 1), (2, 3))]
    pn, pd, rn, rd = e2edutch.metrics.ceafe(clusters, clusters)
    assert math.isclose(pn/pd, 1)

    clusters, gold_clusters, _ = example_clusters()
    pn, pd, rn, rd = e2edutch.metrics.ceafe(clusters, gold_clusters)
    assert math.isclose(pn/pd, 0.43, abs_tol=0.005)



import numpy as np
import torch
from transformers import RobertaTokenizer, RobertaModel, BertTokenizer
from transformers import BertModel, BertForPreTraining, BertConfig


def load_bert(model_name):
    if model_name == 'robbert':
        tokenizer = RobertaTokenizer.from_pretrained("pdelobelle/robBERT-base")
        model = RobertaModel.from_pretrained("pdelobelle/robBERT-base")
    elif model_name == 'bertje':
        tokenizer = BertTokenizer.from_pretrained(
            "wietsedv/bert-base-dutch-cased")
        model = BertModel.from_pretrained("wietsedv/bert-base-dutch-cased")
    elif model_name == 'bert-nl':
        tokenizer = BertTokenizer.from_pretrained("data/bert-nl")
        config = BertConfig.from_json_file("data/bert-nl/config.json")
        model = BertForPreTraining(config).bert
    else:
        raise ValueError('invalid model name')
    model.eval()
    return tokenizer, model


def encode_sentences(sentences, tokenizer, model):
    # Use BERT tokenizer
    sentences_tokenized = [
        [tokenizer.tokenize(word) for word in sentence]
        for sentence in sentences]
    sentences_tokenized_flat = [
        [tok for word in sentence for tok in word]
        for sentence in sentences_tokenized]
    indices_flat = [[i for i, word in enumerate(
        sentence) for tok in word] for sentence in sentences_tokenized]

    max_nrtokens = max(len(s) for s in sentences_tokenized_flat)
    indexed_tokens = np.zeros((len(sentences), max_nrtokens), dtype=int)
    for i, sent in enumerate(sentences_tokenized_flat):
        idx = tokenizer.convert_tokens_to_ids(sent)
        indexed_tokens[i, :len(idx)] = np.array(idx)

    # Convert inputs to PyTorch tensors
    tokens_tensor = torch.tensor(indexed_tokens)
    with torch.no_grad():
        # torch tensor of shape
        # (nr_sentences, sequence_length, hidden_size=768
        bert_output, _ = model(tokens_tensor)

    # Add up tensors for subtokens coming from same word
    max_sentence_length = max(len(s) for s in sentences)
    bert_final = torch.tensor(np.zeros((bert_output.shape[0],
                                        max_sentence_length,
                                        bert_output.shape[2])))
    for sent_id in range(len(sentences)):
        for tok_id, word_id in enumerate(indices_flat[sent_id]):
            bert_final[sent_id, word_id, :] += bert_output[sent_id, tok_id, :]
    bert_final = np.array(bert_final)
    # Add extra axis
    bert_final = np.expand_dims(bert_final, axis=3)
    return bert_final



import e2edutch.conll
import os
__here__ = os.path.dirname(os.path.realpath(__file__))


def test_get_doc_key():
    doc_key = e2edutch.conll.get_doc_key('test', '1')
    assert doc_key == 'test.p.1'


def test_get_doc_key_nopart():
    doc_key = e2edutch.conll.get_doc_key('test')
    assert doc_key == 'test'


def test_get_prediction_map():
    predictions = {'doc1': [[(0, 1)]]}
    prediction_map = e2edutch.conll.get_prediction_map(predictions)
    assert len(prediction_map) == 1
    assert len(prediction_map['doc1']) == 3
    start_map, end_map, word_map = prediction_map['doc1']
    assert start_map[0] == [0]
    assert end_map[1] == [0]
    assert word_map == {}


def test_predictions_to_brackets():
    sentences = [['Een', 'zin', '.'],
                 ['Nog', 'een', 'zin']]
    predictions = [[(1, 1)], [(4, 5)]]
    brackets = e2edutch.conll.clusters_to_brackets(sentences, predictions)
    assert len(brackets) == 2
    assert len(brackets[0]) == 3
    assert brackets[0][1] == '(0)'
    assert brackets[1][1] == '(1'
    assert brackets[1][2] == '1)'


def test_output_conll():
    output_file = '/tmp/tmp.conll'
    sentences = {'doc1.p.1': [['Dit', 'is', 'een', 'test', '.']]}
    predictions = {'doc1.p.1': [[(0, 0), (2, 3)]]}
    with open(output_file, 'w') as fout:
        e2edutch.conll.output_conll(fout, sentences, predictions)
    assert os.path.exists(output_file)
    content = open(output_file).readlines()
    nonempty = [line for line in content if line.strip() != '']
    assert len(nonempty) == 7
    assert content[0].strip() == '#begin document (doc1); part 1'


def test_output_conll_align():
    input_file = os.path.join(__here__, 'data', 'test.conll')
    output_file = '/tmp/tmp.conll'
    predictions = {'doc1.p.1': [[(0, 0), (2, 3)]]}
    with open(input_file) as fin:
        with open(output_file, 'w') as fout:
            e2edutch.conll.output_conll_align(fin, fout, predictions)
    content = open(output_file).readlines()
    nonempty = [line for line in content if line.strip() != '']
    assert len(nonempty) == 7
    assert content[0].strip() == '#begin document (doc1); part 1'



import os
from pathlib import Path
import pyhocon
import errno
import codecs
import collections
import shutil
import logging
import pkg_resources

import numpy as np
import tensorflow.compat.v1 as tf
tf.disable_v2_behavior()

logger = logging.getLogger('e2edutch')


def initialize_from_env(model_name='final', cfg_file=None, model_cfg_file=None):
    '''Read configuration files

    Read configuration files cfg_file and model_cfg_file from provided
    filenames. If none given, use default config files provided by e2edutch:
        cfg/defaults.conf for cfg_file, and
        cfg/models.conf for model_cfg_file


    Configure Tensorflow to use a gpu or cpu based on the environment values of GPU.

    Returns a config dict
    '''
    if "GPU" in os.environ:
        set_gpus(int(os.environ["GPU"]))
    else:
        set_gpus()

    logger.info('Running model: {}'.format(model_name))

    if cfg_file is None:
        cfg_file = pkg_resources.resource_filename(
            "e2edutch", 'cfg/defaults.conf')
    if model_cfg_file is None:
        model_cfg_file = pkg_resources.resource_filename(
            "e2edutch", 'cfg/models.conf')
    config_base = pyhocon.ConfigFactory.parse_file(cfg_file)
    config_model = pyhocon.ConfigFactory.parse_file(model_cfg_file)[model_name]
    config = pyhocon.ConfigTree.merge_configs(config_model, config_base)

    # Override datapath from environment, if set
    if os.environ.get('E2E_HOME', None) is not None:
        config['datapath'] = os.environ['E2E_HOME']

    # Finally, provide fallback for datapath
    if config.get('datapath', None) is None:
        config['datapath'] = Path(__file__).parent / "data"

    config['log_root'] = config['datapath']
    config['log_dir'] = model_name

    mkdirs(os.path.join(config['log_root'], config['log_dir']))

    logger.debug(pyhocon.HOCONConverter.convert(config, 'hocon'))
    return config


def copy_checkpoint(source, target):
    for ext in (".index", ".data-00000-of-00001"):
        shutil.copyfile(source + ext, target + ext)


def make_summary(value_dict):
    return tf.Summary(value=[tf.Summary.Value(
        tag=k, simple_value=v) for k, v in value_dict.items()])


def create_example(text, doc_key='example'):
    import nltk
    nltk.download("punkt")
    from nltk.tokenize import sent_tokenize, word_tokenize
    raw_sentences = sent_tokenize(text)
    sentences = [word_tokenize(s, language='dutch') for s in raw_sentences]
    return {
        "doc_key": doc_key,
        "clusters": [],
        "sentences": sentences
    }


def flatten(l):
    return [item for sublist in l for item in sublist]


def set_gpus(*gpus):
    os.environ["CUDA_VISIBLE_DEVICES"] = ",".join(str(g) for g in gpus)
    logger.info("Setting CUDA_VISIBLE_DEVICES to: {}".format(
        os.environ["CUDA_VISIBLE_DEVICES"]))
    for gpu in tf.config.experimental.list_physical_devices('GPU'):
        tf.config.experimental.set_memory_growth(gpu, True)


def mkdirs(path):
    try:
        os.makedirs(path)
    except OSError as exception:
        if exception.errno != errno.EEXIST:
            raise
    return path


def load_char_dict(char_vocab_path):
    vocab = [u"<unk>"]
    with codecs.open(char_vocab_path, encoding="utf-8") as f:
        vocab.extend(l.strip() for l in f.readlines())
    char_dict = collections.defaultdict(int)
    char_dict.update({c: i for i, c in enumerate(vocab)})
    return char_dict


def maybe_divide(x, y):
    return 0 if y == 0 else x / float(y)


def projection(inputs, output_size, initializer=None):
    return ffnn(inputs, 0, -1, output_size, dropout=None,
                output_weights_initializer=initializer)


def highway(inputs, num_layers, dropout):
    for i in range(num_layers):
        with tf.variable_scope("highway_{}".format(i)):
            j, f = tf.split(projection(inputs, 2 * shape(inputs, -1)), 2, -1)
            f = tf.sigmoid(f)
            j = tf.nn.relu(j)
            if dropout is not None:
                j = tf.nn.dropout(j, dropout)
            inputs = f * j + (1 - f) * inputs
    return inputs


def shape(x, dim):
    return x.get_shape()[dim].value or tf.shape(x)[dim]


def ffnn(inputs, num_hidden_layers, hidden_size, output_size,
         dropout, output_weights_initializer=None):
    if len(inputs.get_shape()) > 3:
        raise ValueError("FFNN with rank {} not supported".format(
            len(inputs.get_shape())))

    if len(inputs.get_shape()) == 3:
        batch_size = shape(inputs, 0)
        seqlen = shape(inputs, 1)
        emb_size = shape(inputs, 2)
        current_inputs = tf.reshape(inputs, [batch_size * seqlen, emb_size])
    else:
        current_inputs = inputs

    for i in range(num_hidden_layers):
        hidden_weights = tf.get_variable("hidden_weights_{}".format(i),
                                         [shape(current_inputs, 1),
                                          hidden_size])
        hidden_bias = tf.get_variable(
            "hidden_bias_{}".format(i), [hidden_size])
        current_outputs = tf.nn.relu(tf.nn.xw_plus_b(
            current_inputs, hidden_weights, hidden_bias))

        if dropout is not None:
            current_outputs = tf.nn.dropout(current_outputs, dropout)
        current_inputs = current_outputs

    output_weights = tf.get_variable("output_weights", [shape(
        current_inputs, 1), output_size],
        initializer=output_weights_initializer)
    output_bias = tf.get_variable("output_bias", [output_size])
    outputs = tf.nn.xw_plus_b(current_inputs, output_weights, output_bias)

    if len(inputs.get_shape()) == 3:
        outputs = tf.reshape(outputs, [batch_size, seqlen, output_size])
    return outputs


def cnn(inputs, filter_sizes, num_filters):
    # num_words = shape(inputs, 0)
    # num_chars = shape(inputs, 1)
    input_size = shape(inputs, 2)
    outputs = []
    for i, filter_size in enumerate(filter_sizes):
        with tf.variable_scope("conv_{}".format(i)):
            w = tf.get_variable("w", [filter_size, input_size, num_filters])
            b = tf.get_variable("b", [num_filters])
        # [num_words, num_chars - filter_size, num_filters]
        conv = tf.nn.conv1d(inputs, w, stride=1, padding="VALID")
        # [num_words, num_chars - filter_size, num_filters]
        h = tf.nn.relu(tf.nn.bias_add(conv, b))
        pooled = tf.reduce_max(h, 1)  # [num_words, num_filters]
        outputs.append(pooled)
    # [num_words, num_filters * len(filter_sizes)]
    return tf.concat(outputs, 1)


def batch_gather(emb, indices):
    batch_size = shape(emb, 0)
    seqlen = shape(emb, 1)
    if len(emb.get_shape()) > 2:
        emb_size = shape(emb, 2)
    else:
        emb_size = 1
    # [batch_size * seqlen, emb]
    flattened_emb = tf.reshape(emb, [batch_size * seqlen, emb_size])
    offset = tf.expand_dims(tf.range(batch_size)
                            * seqlen, 1)  # [batch_size, 1]
    # [batch_size, num_indices, emb]
    gathered = tf.gather(flattened_emb, indices + offset)
    if len(emb.get_shape()) == 2:
        gathered = tf.squeeze(gathered, 2)  # [batch_size, num_indices]
    return gathered


class RetrievalEvaluator(object):
    def __init__(self):
        self._num_correct = 0
        self._num_gold = 0
        self._num_predicted = 0

    def update(self, gold_set, predicted_set):
        self._num_correct += len(gold_set & predicted_set)
        self._num_gold += len(gold_set)
        self._num_predicted += len(predicted_set)

    def recall(self):
        return maybe_divide(self._num_correct, self._num_gold)

    def precision(self):
        return maybe_divide(self._num_correct, self._num_predicted)

    def metrics(self):
        recall = self.recall()
        precision = self.precision()
        f1 = maybe_divide(2 * recall * precision, precision + recall)
        return recall, precision, f1


class EmbeddingDictionary(object):
    def __init__(self, info, datapath='', normalize=True, maybe_cache=None):
        self._size = info["size"]
        self._normalize = normalize
        self._path = os.path.join(datapath, info["path"])
        self._cased = info["cased"] if "cased" in info else True
        if maybe_cache is not None and maybe_cache._path == self._path:
            assert self._size == maybe_cache._size
            self._embeddings = maybe_cache._embeddings
        else:
            self._embeddings = self.load_embedding_dict(self._path)

    @property
    def size(self):
        return self._size

    def load_embedding_dict(self, path):
        logger.info("Loading word embeddings from {}...".format(path))
        default_embedding = np.zeros(self.size)
        embedding_dict = collections.defaultdict(lambda: default_embedding)
        if len(path) > 0:
            vocab_size = None
            with open(path) as f:
                for i, line in enumerate(f.readlines()):
                    word_end = line.find(" ")
                    word = line[:word_end]
                    embedding = np.fromstring(
                        line[word_end + 1:], np.float32, sep=" ")
                    assert len(embedding) == self.size
                    embedding_dict[word] = embedding
            if vocab_size is not None:
                assert vocab_size == len(embedding_dict)
            logger.info("Done loading word embeddings.")
        return embedding_dict

    def __getitem__(self, key):
        if not self._cased:
            key = key.lower()
        embedding = self._embeddings[key]
        if self._normalize:
            embedding = self.normalize(embedding)
        return embedding

    def normalize(self, v):
        norm = np.linalg.norm(v)
        if norm > 0:
            return v / norm
        else:
            return v


class CustomLSTMCell(tf.nn.rnn_cell.RNNCell):
    def __init__(self, num_units, batch_size, dropout):
        self._num_units = num_units
        self._dropout = dropout
        self._dropout_mask = tf.nn.dropout(
            tf.ones([batch_size, self.output_size]), dropout)
        self._initializer = self._block_orthonormal_initializer(
            [self.output_size] * 3)
        initial_cell_state = tf.get_variable(
            "lstm_initial_cell_state", [1, self.output_size])
        initial_hidden_state = tf.get_variable(
            "lstm_initial_hidden_state", [1, self.output_size])
        self._initial_state = tf.nn.rnn_cell.LSTMStateTuple(
            initial_cell_state, initial_hidden_state)

    @property
    def state_size(self):
        return tf.nn.rnn_cell.LSTMStateTuple(
            self.output_size, self.output_size)

    @property
    def output_size(self):
        return self._num_units

    @property
    def initial_state(self):
        return self._initial_state

    def __call__(self, inputs, state, scope=None):
        """Long short-term memory cell (LSTM)."""
        with tf.variable_scope(scope or type(self).__name__):  # CustomLSTMCell
            c, h = state
            h *= self._dropout_mask
            concat = projection(
                tf.concat([inputs, h], 1), 3 * self.output_size,
                initializer=self._initializer)
            i, j, o = tf.split(concat, num_or_size_splits=3, axis=1)
            i = tf.sigmoid(i)
            new_c = (1 - i) * c + i * tf.tanh(j)
            new_h = tf.tanh(new_c) * tf.sigmoid(o)
            new_state = tf.nn.rnn_cell.LSTMStateTuple(new_c, new_h)
            return new_h, new_state

    def _orthonormal_initializer(self, scale=1.0):
        def _initializer(shape, dtype=tf.float32, partition_info=None):
            M1 = np.random.randn(shape[0], shape[0]).astype(np.float32)
            M2 = np.random.randn(shape[1], shape[1]).astype(np.float32)
            Q1, R1 = np.linalg.qr(M1)
            Q2, R2 = np.linalg.qr(M2)
            Q1 = Q1 * np.sign(np.diag(R1))
            Q2 = Q2 * np.sign(np.diag(R2))
            n_min = min(shape[0], shape[1])
            params = np.dot(Q1[:, :n_min], Q2[:n_min, :]) * scale
            return params
        return _initializer

    def _block_orthonormal_initializer(self, output_sizes):
        def _initializer(shape, dtype=np.float32, partition_info=None):
            assert len(shape) == 2
            assert sum(output_sizes) == shape[1]
            initializer = self._orthonormal_initializer()
            params = np.concatenate(
                [initializer([shape[0], o], dtype, partition_info) for o in output_sizes], 1)
            return params
        return _initializer



import re
import tempfile
import subprocess
import operator
import collections

BEGIN_DOCUMENT_REGEX = re.compile(
    r"#begin document \(?([^\);]*)\)?;?(?: part (\d+))?")
COREF_RESULTS_REGEX = re.compile(
    r".*Coreference: Recall: \([0-9.]+ / [0-9.]+\) ([0-9.]+)%\tPrecision: \([0-9.]+ / [0-9.]+\) ([0-9.]+)%\tF1: ([0-9.]+)%.*", re.DOTALL)


def get_doc_key(doc_id, part=None):
    if part is None:
        return doc_id
    else:
        return '{}.p.{}'.format(doc_id, part)


def get_reverse_doc_key(doc_key):
    segments = doc_key.split('.p.')
    if len(segments) > 1:
        part = segments[-1]
        doc_id = '.p.'.join(segments[:-1])
    else:
        doc_id = doc_key
        part = None
    return doc_id, part


def get_prediction_map(predictions):
    prediction_map = {}
    for doc_key, clusters in predictions.items():
        start_map = collections.defaultdict(list)
        end_map = collections.defaultdict(list)
        word_map = collections.defaultdict(list)
        for cluster_id, mentions in enumerate(clusters):
            for start, end in mentions:
                if start == end:
                    word_map[start].append(cluster_id)
                else:
                    start_map[start].append((cluster_id, end))
                    end_map[end].append((cluster_id, start))
        for k, v in start_map.items():
            start_map[k] = [cluster_id for cluster_id, end in sorted(
                v, key=operator.itemgetter(1), reverse=True)]
        for k, v in end_map.items():
            end_map[k] = [cluster_id for cluster_id, start in sorted(
                v, key=operator.itemgetter(1), reverse=True)]
        prediction_map[doc_key] = (start_map, end_map, word_map)
    return prediction_map


def clusters_to_brackets(sentences, predictions):
    prediction_map = get_prediction_map({'': predictions})
    start_map, end_map, word_map = prediction_map['']
    word_index = 0
    brackets_list = []
    for sent in sentences:
        sent_brackets_list = []
        for i, word in enumerate(sent):
            coref_list = []
            if word_index in end_map:
                for cluster_id in end_map[word_index]:
                    coref_list.append("{})".format(cluster_id))
            if word_index in word_map:
                for cluster_id in word_map[word_index]:
                    coref_list.append("({})".format(cluster_id))
            if word_index in start_map:
                for cluster_id in start_map[word_index]:
                    coref_list.append("({}".format(cluster_id))
            coref = '-' if len(coref_list) == 0 else "|".join(coref_list)
            sent_brackets_list.append(coref)
            word_index += 1
        brackets_list.append(sent_brackets_list)
    return brackets_list


def output_conll(output_file, sentences, predictions):
    """
    Output the tokens and coreferences in CONLL-2012 format

    Args:
        output_file (File or IOBase): File to write the CONLL to
        sentences (dict): keys are the doc_keys, values are the sentences of
                          that doc
        predictions (dict): keys are the doc_keys, values are the predicted
                            clusters of that doc
    """
    for doc_key in sentences:
        brackets = clusters_to_brackets(sentences[doc_key], predictions[doc_key])
        doc_id, part = get_reverse_doc_key(doc_key)
        if part is None:
            output_file.write("#begin document ({});\n\n".format(doc_id))
        else:
            output_file.write(
                "#begin document ({}); part {}\n\n".format(
                    doc_id, part))
        for sent, brack_sent in zip(sentences[doc_key], brackets):
            for i, word in enumerate(sent):
                coref = brack_sent[i]
                line = '\t'.join([doc_id, str(i), word, coref])
                output_file.write(line + '\n')
            output_file.write('\n')
        output_file.write('#end document\n')


def output_conll_align(input_file, output_file, predictions):
    prediction_map = get_prediction_map(predictions)

    word_index = 0
    for line in input_file.readlines():
        row = line.split()
        if len(row) == 0:
            output_file.write("\n")
        elif row[0].startswith("#"):
            begin_match = re.match(BEGIN_DOCUMENT_REGEX, line)
            if begin_match:
                doc_key = get_doc_key(*begin_match.groups())
                start_map, end_map, word_map = prediction_map[doc_key]
                word_index = 0
            output_file.write(line)
            output_file.write("\n")
        else:
            coref_list = []
            if word_index in end_map:
                for cluster_id in end_map[word_index]:
                    coref_list.append("{})".format(cluster_id))
            if word_index in word_map:
                for cluster_id in word_map[word_index]:
                    coref_list.append("({})".format(cluster_id))
            if word_index in start_map:
                for cluster_id in start_map[word_index]:
                    coref_list.append("({}".format(cluster_id))

            if len(coref_list) == 0:
                row[-1] = "-"
            else:
                row[-1] = "|".join(coref_list)

            output_file.write("   ".join(row))
            output_file.write("\n")
            word_index += 1


def official_conll_eval(gold_path, predicted_path,
                        metric, official_stdout=False):
    cmd = ["conll-2012/scorer/v8.01/scorer.pl",
           metric, gold_path, predicted_path, "none"]
    process = subprocess.Popen(cmd, stdout=subprocess.PIPE)
    stdout, stderr = process.communicate()
    process.wait()

    stdout = stdout.decode("utf-8")
    if stderr is not None:
        print(stderr)

    if official_stdout:
        print("Official result for {}".format(metric))
        print(stdout)

    coref_results_match = re.match(COREF_RESULTS_REGEX, stdout)
    recall = float(coref_results_match.group(1))
    precision = float(coref_results_match.group(2))
    f1 = float(coref_results_match.group(3))
    return {"r": recall, "p": precision, "f": f1}


def evaluate_conll(gold_path, predictions, official_stdout=False):
    with tempfile.NamedTemporaryFile(delete=False, mode="w") as pred_file:
        with open(gold_path, "r") as gold_file:
            output_conll_align(gold_file, pred_file, predictions)
        print("Predicted conll file: {}".format(pred_file.name))
    return {m: official_conll_eval(
        gold_file.name, pred_file.name, m, official_stdout)
        for m in ("muc", "bcub", "ceafe")}



from . import __version__
import logging

import KafNafParserPy
from KafNafParserPy import KafNafParser
from lxml.etree import XMLSyntaxError
import itertools
import tensorflow.compat.v1 as tf
tf.disable_v2_behavior()

logger = logging.getLogger('e2edutch')
this_name = 'Coreference resolution based on e2e model'


def get_naf(input_filename):
    try:
        naf = KafNafParser(input_filename)
    except XMLSyntaxError:
        with open(input_filename) as input_file:
            input = input_file.read()
        if "<NAF" in input and "</NAF>" in input:
            # I'm guessing this should be a NAF file but something is wrong
            logger.exception("Error parsing NAF file")
            raise
        naf = KafNafParser(type="NAF")
        naf.set_version("3.0")
        naf.set_language("nl")
        naf.lang = "nl"
        naf.raw = input
        naf.set_raw(naf.raw)
    return naf


def get_naf_from_sentences(sentences):
    naf_obj = KafNafParser(type="NAF")
    naf_obj.set_version("3.0")
    naf_obj.set_language("nl")
    naf_obj.lang = "nl"
    naf_obj.raw = '\n'.join([' '.join(s) for s in sentences])
    naf_obj.set_raw(naf_obj.raw)
    # Create text layer
    wcount = 1
    offsets = {}
    txt = naf_obj.get_raw()
    token_ids = []
    for sid, sentence in enumerate(sentences):
        token_ids_sub = []
        for token in sentence:
            token_obj = KafNafParserPy.Cwf(type=naf_obj.get_type())
            token_id = 'w{}'.format(wcount)
            token_length = len(token)
            offsets[wcount] = txt.find(token, offsets.get(wcount - 1, 0))
            token_obj.set_id(token_id)
            token_obj.set_length(str(token_length))
            # token_obj.set_offset(str(offset)) # Is this correct????
            token_obj.set_para('1')
            token_obj.set_sent(str(sid + 1))
            token_obj.set_text(token)
            token_obj.set_offset(str(offsets[wcount]))
            token_ids_sub.append(token_id)
            wcount += 1
            naf_obj.add_wf(token_obj)
        token_ids.append(token_ids_sub)
    # Create term layers
    term_ids = []
    count_terms = 0
    for sid, (sentence, token_ids_sub) in enumerate(zip(sentences, token_ids)):
        term_ids_sub = []
        logger.info('Creating the term layer...')
        for num_token, (token, token_id) in enumerate(
                zip(sentence, token_ids_sub)):
            new_term_id = 't_' + str(count_terms)
            count_terms += 1
            term_ids_sub.append(new_term_id)
            term_obj = KafNafParserPy.Cterm(type=naf_obj.get_type())
            term_obj.set_id(new_term_id)
            new_span = KafNafParserPy.Cspan()
            new_span.create_from_ids([token_id])
            term_obj.set_span(new_span)
            naf_obj.add_term(term_obj)
        term_ids.append(term_ids_sub)

    return naf_obj, term_ids


def create_coref_layer(knaf_obj, clusters, term_ids):
    term_ids_list = list(itertools.chain.from_iterable(term_ids))
    for cluster_id, cluster in enumerate(clusters):
        coref_obj = KafNafParserPy.Ccoreference(type=knaf_obj.get_type())
        coref_obj.set_id('co{}'.format(cluster_id + 1))
        coref_obj.set_type('entity')
        for start, end in cluster:
            coref_obj.add_span(term_ids_list[start:end + 1])
            span_text = []
            for term_id in term_ids_list[start:end + 1]:
                word_ids = knaf_obj.get_term(term_id).get_span_ids()
                for word_id in word_ids:
                    word = knaf_obj.get_token(word_id).get_text()
                    span_text.append(word)
            span_text = ' '.join(span_text)
            # TODO: output span_text as comment
        knaf_obj.add_coreference(coref_obj)
    return knaf_obj


def add_linguistic_processors(in_obj):
    name = this_name

    my_lp = KafNafParserPy.Clp()
    my_lp.set_name(name)
    my_lp.set_version(__version__)
    my_lp.set_timestamp()
    in_obj.add_linguistic_processor('coreferences', my_lp)

    return in_obj


def get_jsonlines(knaf_obj):
    sent_term_tok = []

    for term in knaf_obj.get_terms():
        for tok_id in term.get_span_ids():
            tok = knaf_obj.get_token(tok_id)
            sent_term_tok.append(
                (tok.get_sent(), term.get_id(), tok_id, tok.get_text()))

    sentences = []
    term_ids = []
    tok_ids = []
    for sent_id, idlist in itertools.groupby(sent_term_tok, lambda t: t[0]):
        idlist = list(idlist)
        sentences.append([t[3] for t in idlist])
        term_ids.append([t[1] for t in idlist])
        tok_ids.append([t[2] for t in idlist])

    jsonlines_obj = {'doc_key': str(knaf_obj.get_filename()),
                     'sentences': sentences,
                     'clusters': []
                     }
    return jsonlines_obj, term_ids, tok_ids



import os
import stanza
import logging

from pathlib import Path

from e2edutch import util
from e2edutch import coref_model as cm
from e2edutch.download import download_data
from e2edutch.predict import Predictor

from stanza.pipeline.processor import Processor, register_processor
from stanza.models.common.doc import Document, Span


# Add a Clusters property to documents as a List of List of Span:
# Clusters is a List of cluster, cluster is a List of Span
def clusterSetter(self, value):
    if isinstance(value, type([])):
        self._clusters = value
    else:
        logger.error('Clusters must be a List')

stanza.models.common.doc.Document.add_property('clusters', default='[]', setter=clusterSetter)


import tensorflow.compat.v1 as tf

logger = logging.getLogger('e2edutch')
logger.setLevel(logging.INFO)
logger.addHandler(logging.StreamHandler())


@register_processor('coref')
class CorefProcessor(Processor):
    ''' Processor that appends coreference information '''
    _requires = set(['tokenize'])
    _provides = set(['coref'])
    
    def __init__(self, config, pipeline, use_gpu):
        # Make e2edutch follow Stanza's GPU settings:
        # set the environment value for GPU, so that initialize_from_env picks it up.
        #if use_gpu:
        #    os.environ['GPU'] = ' '.join(tf.config.experimental.list_physical_devices('GPU'))
        #else:
        #    if 'GPU' in os.environ['GPU'] :
        #        os.environ.pop('GPU')

        self.e2econfig = util.initialize_from_env(model_name='final')

        # Override datapath and log_root:
        # store e2edata with the Stanza resources, ie. a 'stanza_resources/nl/coref' directory
        self.e2econfig['datapath'] = Path(config['model_path']).parent
        self.e2econfig['log_root'] = Path(config['model_path']).parent

        # Download data files if not present
        download_data(self.e2econfig)

        # Start and stop a session to cache all models
        predictor = Predictor(config=self.e2econfig)
        predictor.end_session()

    def _set_up_model(self, *args):
        print ('_set_up_model')
        pass

    def process(self, doc):

        predictor = Predictor(config=self.e2econfig)

        # build the example argument for predict:
        #   example (dict): dict with the following fields:
        #                     sentences ([[str]])
        #                     doc_id (str)
        #                     clusters ([[(int, int)]]) (optional)
        example = {}
        example['sentences'] = []
        example['doc_id'] = 'document_from_stanza'  # TODO check what this should be
        example['doc_key'] = 'undocumented'  # TODO check what this should be

        for sentence in doc.sentences:
            s = []
            for word in sentence.words:
                s.append(word.text)
            example['sentences'].append(s)

        predicted_clusters = predictor.predict(example)  # a list of tuples

        # Add the predicted clusters back to the Stanza document

        clusters = []
        for predicted_cluster in predicted_clusters:  # a tuple of entities
            cluster = []
            for predicted_reference in predicted_cluster:  # a tuple of (start, end) word
                start, end = predicted_reference

                # find the sentence_id of the sentence containing this reference
                sentence_id = 0
                sentence = doc.sentences[0]
                sentence_start_word = 0
                sentence_end_word = len(sentence.words) - 1

                while sentence_end_word < start:
                    sentence_start_word = sentence_end_word + 1

                    # move to the next sentence
                    sentence_id += 1
                    sentence = doc.sentences[sentence_id]

                    sentence_end_word = sentence_start_word + len(sentence.words) - 1

                # start counting words from the start of this sentence
                start -= sentence_start_word
                end -= sentence_start_word

                span = Span(  # a list of Tokens
                        tokens=[word.parent for word in sentence.words[start:end + 1]],
                        doc=doc,
                        type='COREF',
                        sent=doc.sentences[sentence_id]
                        )
                cluster.append(span)

            clusters.append(cluster)

        doc.clusters = clusters

        predictor.end_session()

        return doc



import stanza
import e2edutch.stanza
import tensorflow.compat.v1 as tf


def test_processor():
    nlp = stanza.Pipeline(lang='nl', processors='tokenize,coref')
    text = 'Dit is een tekst.'
    doc = nlp(text)
    # TODO: asserts about the doc having corefs



import e2edutch.predict
from collections import OrderedDict
import tensorflow.compat.v1 as tf


def test_predict():
    example = {
        'doc_key': 'test',
        'sentences': [['Een', 'zin', '.']]
    }
    predictor = e2edutch.predict.Predictor(model_name='final')
    clusters = predictor.predict(example)
    predictor.end_session()
    assert(len(clusters) >= 0)



import setuptools
import setuptools.command.build_py
import os
import subprocess
import pkg_resources

__version__ = None

with open("README.md", "r") as fh:
    long_description = fh.read()

with open(os.path.join(os.path.dirname(__file__), 'e2edutch/__version__.py')) as versionpy:
    exec(versionpy.read())


class BuildPyCommand(setuptools.command.build_py.build_py):
    """Build the tensorflow kernels, and proceed with default build."""

    def run(self):
        try:
            import tensorflow as tf
        except ImportError:
            raise ImportError("This module requires tensorflow to be installed. " +
                              "Install it first with `pip install tensorflow`")
        args = ["g++", "-std=c++11", "-shared"]
        args += [
            pkg_resources.resource_filename("e2edutch", "coref_kernels.cc"),
            "-o",
            pkg_resources.resource_filename("e2edutch", "coref_kernels.so")
        ]
        args += ["-fPIC"]
        args += tf.sysconfig.get_compile_flags() + tf.sysconfig.get_link_flags()
        args += ["-O2",
                 "-D_GLIBCXX_USE_CXX11_ABI=0"]
        subprocess.check_call(args)
        setuptools.command.build_py.build_py.run(self)


setuptools.setup(
    name="e2e-Dutch",
    version=__version__,
    author="Dafne van Kuppevelt",
    author_email="d.vankuppevelt@esciencecenter.nl",
    description="Coreference resolution with e2e for Dutch",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/Filter-Bubble/e2e-Dutch",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "Development Status :: 4 - Beta",
        "Environment :: Console",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
    ],
    include_package_data=True,
    package_data={'e2edutch': ["cfg/*.conf", "coref_kernels.so"]},
    cmdclass={
        "build_py": BuildPyCommand
    },
    test_suite='test',
    python_requires='>=3.6',
    install_requires=[
        "tensorflow>=2.0.0",
        "h5py",
        "nltk",
        "pyhocon",
        "scipy",
        "scikit-learn",
        "torch",
        "transformers<=3.5.1",
        "KafNafParserPy",
        "stanza"
    ],
    tests_require=[
        'pytest',
        'pytest-cov',
        'pycodestyle',
    ],
)



#!/usr/bin/env python
import os
import time
import argparse
import logging
import tensorflow.compat.v1 as tf
tf.disable_v2_behavior()

from e2edutch import util
from e2edutch import coref_model as cm


def get_parser():
    parser = argparse.ArgumentParser()
    parser.add_argument('config')
    parser.add_argument('--train',
                        type=str,
                        default=None,
                        help="jsonlines file used for training")
    parser.add_argument('--eval',
                        type=str,
                        default=None,
                        help="jsonlines file used for evaluating")
    parser.add_argument('--eval_conll',
                        type=str,
                        default=None,
                        help="conll file used for evaluating")
    parser.add_argument('--cfg_file',
                        type=str,
                        default=None,
                        help="config file")
    parser.add_argument('--model_cfg_file',
                        type=str,
                        default=None,
                        help="model config file")
    parser.add_argument('-v', '--verbose', action='store_true')
    return parser


def main(args=None):
    args = get_parser().parse_args()
    if args.verbose:
        logging.basicConfig(level=logging.DEBUG)
    config = util.initialize_from_env(args.config,
                                      args.cfg_file,
                                      args.model_cfg_file)
    # Overwrite train and eval file if specified
    if args.train is not None:
        config['train_path'] = args.train
    if args.eval is not None:
        config['eval_path'] = args.eval
    if args.eval_conll is not None:
        config['conll_eval_path'] = args.eval_conll

    report_frequency = config["report_frequency"]
    eval_frequency = config["eval_frequency"]

    model = cm.CorefModel(config)
    saver = tf.train.Saver()

    log_dir = os.path.join(config['log_root'], config['log_dir'])
    writer = tf.summary.FileWriter(log_dir, flush_secs=20)

    max_f1 = 0

    with tf.Session() as session:
        session.run(tf.global_variables_initializer())
        model.start_enqueue_thread(session)
        accumulated_loss = 0.0

        ckpt = tf.train.get_checkpoint_state(log_dir)
        if ckpt and ckpt.model_checkpoint_path:
            print("Restoring from: {}".format(ckpt.model_checkpoint_path))
            saver.restore(session, ckpt.model_checkpoint_path)

        initial_time = time.time()
        while True:
            tf_loss, tf_global_step, _ = session.run(
                [model.loss, model.global_step, model.train_op])
            accumulated_loss += tf_loss

            if tf_global_step % report_frequency == 0:
                total_time = time.time() - initial_time
                steps_per_second = tf_global_step / total_time

                average_loss = accumulated_loss / report_frequency
                print("[{}] loss={:.2f}, steps/s={:.2f}"
                      .format(tf_global_step,
                              average_loss,
                              steps_per_second))
                writer.add_summary(util.make_summary(
                    {"loss": average_loss}), tf_global_step)
                accumulated_loss = 0.0

            if tf_global_step % eval_frequency == 0:
                saver.save(session, os.path.join(log_dir, "model"),
                           global_step=tf_global_step)
                eval_summary, eval_f1 = model.evaluate(session)

                if eval_f1 > max_f1:
                    max_f1 = eval_f1
                    util.copy_checkpoint(os.path.join(
                        log_dir, "model-{}".format(tf_global_step)),
                                         os.path.join(log_dir, "model.max.ckpt"))

                writer.add_summary(eval_summary, tf_global_step)
                writer.add_summary(util.make_summary(
                    {"max_eval_f1": max_f1}), tf_global_step)

                print("[{}] evaL_f1={:.2f}, max_f1={:.2f}".format(
                    tf_global_step, eval_f1, max_f1))


if __name__ == "__main__":
    main()



import requests
import gzip
import zipfile
import argparse
import logging
from tqdm import tqdm
from pathlib import Path
from e2edutch import util

logger = logging.getLogger()


def download_file(url, path):
    """
    Download a URL into a file as specified by `path`.
    """
    # This function is copied from stanza
    # https://github.com/stanfordnlp/stanza/blob/f0338f891a03e242c7e11e440dec6e191d54ab77/stanza/resources/common.py#L103
    r = requests.get(url, stream=True)
    with open(path, 'wb') as f:
        file_size = int(r.headers.get('content-length'))
        default_chunk_size = 131072
        desc = 'Downloading ' + url
        with tqdm(total=file_size, unit='B', unit_scale=True,
                  desc=desc) as pbar:
            for chunk in r.iter_content(chunk_size=default_chunk_size):
                if chunk:
                    f.write(chunk)
                    f.flush()
                    pbar.update(len(chunk))


def download_data(config={}):
    # Create the data directory if it doesn't exist yet
    data_dir = Path(config['datapath'])
    logger.info('Downloading to {}'.format(data_dir))
    data_dir.mkdir(parents=True, exist_ok=True)

    # Download word vectors
    logger.info('Download word vectors')
    url = "https://dl.fbaipublicfiles.com/fasttext/vectors-crawl/cc.nl.300.vec.gz"
    fname = data_dir / 'fasttext.300.vec'
    fname_gz = data_dir / 'fasttext.300.vec.gz'
    if not fname.exists():
        download_file(url, fname_gz)
        with gzip.open(fname_gz, 'rb') as fin:
            with open(fname, 'wb') as fout:
                # We need to remove the first line
                for i, line in enumerate(fin.readlines()):
                    if i > 0:
                        fout.write(line)
        # Remove gz file
        fname_gz.unlink()
    else:
        logger.info('Word vectors file already exists')

    # Download e2e dutch model_
    logger.info('Download e2e model')
    url = "https://surfdrive.surf.nl/files/index.php/s/UnZMyDrBEFunmQZ/download"
    fname_zip = data_dir / 'model.zip'
    log_dir_name = data_dir / 'final'
    model_file = log_dir_name / 'model.max.ckpt.index'
    if not fname_zip.exists() and not model_file.exists():
        download_file(url, fname_zip)
    if not model_file.exists():
        with zipfile.ZipFile(fname_zip, 'r') as zfile:
            zfile.extractall(data_dir)
        Path(data_dir / 'logs' / 'final').rename(log_dir_name)
        Path(data_dir, 'logs').rmdir()
    else:
        logger.info('E2e model file already exists')

    # Download char_dict
    logger.info('Download char dict')
    url = "https://github.com/Filter-Bubble/e2e-Dutch/raw/v0.2.0/data/char_vocab.dutch.txt"
    fname = data_dir / 'char_vocab.dutch.txt'
    if not fname.exists():
        download_file(url, fname)
    else:
        logger.info('Char dict file already exists')


def get_parser():
    parser = argparse.ArgumentParser()
    parser.add_argument('-d', '--datapath', default=None)
    parser.add_argument('-v', '--verbose', action='store_true')
    return parser


def main():
    parser = get_parser()
    args = parser.parse_args()
    if args.verbose:
        # logger.setLevel(logging.INFO)
        logging.basicConfig(level=logging.INFO)
    # To do: argparse for config file
    if args.datapath is None:
        config = util.initialize_from_env(model_name='final')
    else:
        config = {'datapath': args.datapath}
    download_data(config)


if __name__ == "__main__":
    main()
